package cz.zcu.kiv.comav.loaders.sofa2;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.WildcardFileFilter;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import cz.zcu.kiv.comav.core.export.loaders.ComponentLoader;
import cz.zcu.kiv.comav.entmm.Component;
import cz.zcu.kiv.comav.entmm.ComponentType;
import cz.zcu.kiv.comav.entmm.ENTMMFactory;
import cz.zcu.kiv.comav.entmm.Element;
import cz.zcu.kiv.comav.entmm.ElementElementBinding;
import cz.zcu.kiv.comav.entmm.Tag;
import cz.zcu.kiv.comav.entmm.TagDefinition;
import cz.zcu.kiv.comav.entmm.Trait;
import cz.zcu.kiv.comav.entmm.TraitDefinition;

/**
 * This class do all work around loading SOFA2 sofatypes in entmm data structure
 * 
 * It has to contain all the data mapping directly in the code
 * 
 * @author Martin Sloup
 * 
 */
public class Sofa2LocalComponentLoader extends ComponentLoader {
	private final static Logger logger = Logger
			.getLogger(Sofa2LocalComponentLoader.class);

	private static final String SOFATYPE_SCHEME = "sofatype";

	// Common elements
	private static final String ELEMENT_PROPERTY = "property";
	private static final String ATTRIBUTE_NAME = "name";
	private static final String ATTRIBUTE_TYPE = "type";

	// Assembly
	private static final String ELEMENT_ASSEMBLY = "assembly";
	private static final String ELEMENT_SUBCOMPONENT = "subcomponent";
	private static final String ATTRIBUTE_TOP_LEVEL_ARCH = "top-level-arch";
	private static final String ATTRIBUTE_ARCH = "arch";

	// Architecture
	private static final String ELEMENT_ARCHITECTURE = "architecture";
	private static final String ATTRIBUTE_FRAME = "frame";
	private static final String ATTRIBUTE_IMPL = "impl";

	// Frame
	private static final String ELEMENT_PROVIDES = "provides";
	private static final String ELEMENT_PROVIDED_INTERFACE = "providedInterface";
	private static final String ELEMENT_REQUIRES = "requires";
	private static final String ELEMENT_REQUIRED_INTERFACE = "requiredInterface";
	private static final String ATTRIBUTE_ITF_TYPE = "itf-type";
	private static final String ATTRIBUTE_COMM_STYLE = "comm-style";

	// Traits
	private static final String TRAIT_PROVIDED_INTERFACE = "Provided interface";
	private static final String TRAIT_REQUIRED_INTERFACE = "Required interface";
	private static final String TRAIT_PROPERTIES = "Properties";

	// Tags
	private static final String TAG_FRAME = "frame";
	private static final String TAG_IMPLEMENTATION = "implementation";
	private static final String TAG_COMMUNICATION_STYLE = "Communication style";
	private static final String TAG_INTERFACE_TYPE = "Interface type";
	private static final String TAG_VERSION = "Version";

	/**
	 * Path to osgi model specification
	 */
	private static final String SOFA2_MODEL_PATH = "/cz/zcu/kiv/comav/loaders/sofa2/SOFA2.entmm";
	private static final String ADL_FILENAME = "adl.xml";
	
	private Hashtable<URI, File> adls;
	private ArrayList<URI> assemblies;

	private File basePath = null;

	/**
	 * SOFA2 component model and attribute mapping is done here
	 */
	public Sofa2LocalComponentLoader(File basePath) {
		super(SOFA2_MODEL_PATH);
		this.basePath = basePath;
		this.adls = new Hashtable<URI, File>();
		this.assemblies = new ArrayList<URI>();
		sniffForAdls();
	}
	
	private void sniffForAdls() {
		Collection<File> files = FileUtils.listFiles(basePath, new WildcardFileFilter(ADL_FILENAME), DirectoryFileFilter.INSTANCE);
		for (File file : files) {
			URI uri = getURIFromFile(file);
			if (uri != null)
				adls.put(uri, file);
		}
	}
	
	private URI getURIFromFile(File adl) {
		URI ret = null;
		
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			Document doc = builder.parse(adl);
			org.w3c.dom.Element tree = doc.getDocumentElement();
			ret = URI.create(SOFATYPE_SCHEME + "://" + tree.getAttribute(ATTRIBUTE_NAME));
			if (ELEMENT_ASSEMBLY.equals(tree.getNodeName()))
				assemblies.add(ret);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return ret;
	}

	@Override
	public void load(URI[] components) throws IOException {
		// SOFA2 has only one component type - architecture
		ComponentType ct = model.getComponentSet()[0];

		for (URI component : components) {
			if (component == null)
				component = loadAssemblyFromBasePath();

			if (!SOFATYPE_SCHEME.equals(component.getScheme()))
				throw new IllegalArgumentException(
						"SOFATYPE scheme in URI is supported only.");

			Component comp = loadAssembly(component, ct);
			if (comp != null) {
				componentList.add(comp);
			}
		}
		createBindings(componentList, factory);
	}
	
	private URI loadAssemblyFromBasePath() {
		URI assembly = null;
		
		assembly = URI.create("sofatype://org.objectweb.dsrg.sofa.examples.logdemo.assm.LogDemo");
		
		return assembly;
	}

	private void createBindings(List<Component> componentList,
			ENTMMFactory factory) {
		for (Component localComponent : componentList) {
			createBinding(localComponent, componentList,
					TRAIT_REQUIRED_INTERFACE, TRAIT_PROVIDED_INTERFACE, factory);

			if (localComponent.getComponentSetLength() > 0) {
				createBindings(localComponent.getComponentSetList(), factory);
			}
		}
	}

	private void createBinding(Component localComponent,
			List<Component> aliedComponens, String localTraitName,
			String aliedTraitName, ENTMMFactory factory) {
		for (Trait localTrait : localComponent.getTraitSet()) {
			if (localTrait.getDef().getName().equals(localTraitName)) {
				for (Element localElement : localTrait.getElementSet()) {
					for (Component aliedComponent : aliedComponens) {
						// for(Tag aliedTag : aliedComponent.getTagSet()) {
						// createComponentElementBinding(localComponent,
						// aliedComponent, localTraitName, localElement,
						// factory);
						// }
						for (Trait aliedTrait : aliedComponent.getTraitSet()) {
							if (aliedTrait.getDef().getName()
									.equals(aliedTraitName)) {
								for (Element aliedElement : aliedTrait
										.getElementSet()) {
									if (localElement.getName().equals(
											aliedElement.getName())) {
										// TRAIT_REQUIRED_INTERFACE and
										// TRAIT_PROVIDED_INTERFACE must have
										// same value for TAG_INTERFACE_TYPE
										if (TRAIT_REQUIRED_INTERFACE
												.equals(localTraitName)
												&& !hasSameInterfaceType(
														localElement,
														aliedElement))
											continue;
										createElementElementBinding(
												localComponent, aliedComponent,
												localTraitName, localElement,
												aliedElement, factory);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	private boolean hasSameInterfaceType(Element localElement,
			Element aliedElement) {
		for (Tag localTag : localElement.getTagsetList()) {
			for (Tag aliedTag : aliedElement.getTagsetList()) {
				if (TAG_INTERFACE_TYPE.equals(localTag.getDef().getName())
						&& TAG_INTERFACE_TYPE.equals(aliedTag.getDef()
								.getName()) && localTag.getValue() != null
						&& localTag.getValue().equals(aliedTag.getValue())) {
					return true;
				}
			}
		}
		return false;
	}

	private void createElementElementBinding(Component localComponent,
			Component aliedComponent, String localTraitName,
			Element localElement, Element aliedElement, ENTMMFactory factory) {
		ElementElementBinding localBinding = factory
				.createElementElementBinding();
		localBinding.setLocal(localElement);
		localBinding.setAlien(aliedElement);
		localBinding.setProvided(false);
		localComponent.getBindingsList().add(localBinding);

		ElementElementBinding aliedBinding = factory
				.createElementElementBinding();
		aliedBinding.setLocal(aliedElement);
		aliedBinding.setAlien(localElement);
		aliedBinding.setProvided(true);
		aliedComponent.getBindingsList().add(aliedBinding);
	}

	/**
	 * Loads an assembly XML file and process data mining
	 * 
	 * @param assemblyName
	 *            Assembly name with sofatype or sofarep protocol
	 * @param ct
	 *            Architecture {@link ComponentType}
	 * @param factory
	 *            ENTMM factory object
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 */
	private Component loadAssembly(URI assemblyURI, ComponentType ct)
			throws IOException {
		Component comp = null;

		try {
			InputStream is = new FileInputStream(adls.get(assemblyURI));
			XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
			XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

			while (xmlReader.hasNext()) {
				xmlReader.next();

				if (xmlReader.isStartElement()
						&& ELEMENT_ASSEMBLY.equals(xmlReader.getLocalName())) {
					String architectureName = xmlReader.getAttributeValue(null,
							ATTRIBUTE_TOP_LEVEL_ARCH);

					// sofarep xml not use this attribute. Search for first
					// Architecture element.
					if (architectureName == null)
						continue;

					// map architecture URI to filesystem
					URI architectureURI = URI.create(architectureName);

					comp = loadArchitecture(architectureURI, ct);
					xmlReader.next();
				}

				// parse subcomponent
				if (xmlReader.isStartElement()
						&& ELEMENT_SUBCOMPONENT
								.equals(xmlReader.getLocalName())) {
					comp.getComponentSetList().add(
							loadSubComponent(xmlReader, assemblyURI, ct));
				}
			}

			is.close();
		} catch (XMLStreamException e) {
			throw new IOException("Error while parsing XML file", e);
		}
		return comp;
	}

	/**
	 * Process a subcomponents from Assembly XML
	 * 
	 * @param xmlReader
	 *            {@link XMLStreamReader} object
	 * @param ct
	 *            Architecture {@link ComponentType}
	 * @param factory
	 *            ENTMM factory object
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private Component loadSubComponent(XMLStreamReader xmlReader,
			URI assemblyURI, ComponentType ct) throws IOException,
			XMLStreamException {
		String architectureName = xmlReader.getAttributeValue(null,
				ATTRIBUTE_ARCH);

		Component comp = null;

		if (architectureName != null) {
			// map architecture URI to filesystem
			URI architectureURI = URI.create(architectureName);

			comp = loadArchitecture(architectureURI, ct);
		}

		if (!xmlReader.isEndElement()) {
			while (!xmlReader.isEndElement()
					|| !ELEMENT_SUBCOMPONENT.equals(xmlReader.getLocalName())) {
				xmlReader.next();

				// parse subcomponent
				if (comp != null
						&& xmlReader.isStartElement()
						&& ELEMENT_SUBCOMPONENT
								.equals(xmlReader.getLocalName())) {
					comp.getComponentSetList().add(
							loadSubComponent(xmlReader, assemblyURI, ct));
				}
			}
		}

		return comp;
	}

	/**
	 * Loads an architecture XML file and process data mining
	 * 
	 * @param architectureName
	 *            Architecture name with sofatype protocol
	 * @param ct
	 *            Architecture {@link ComponentType}
	 * @param factory
	 *            ENTMM factory object
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private Component loadArchitecture(URI architectureURI, ComponentType ct)
			throws IOException, XMLStreamException {
		Component comp = factory.createComponent();
		comp.setDef(ct);

		InputStream is = new FileInputStream(adls.get(architectureURI));

		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

		Properties properties = new Properties();
		TraitDefinition traitDefinitionProperties = null;
		// store trait definition
		for (TraitDefinition traitDefinition : ct.getTraitSetList()) {
			if (TRAIT_PROPERTIES.equals(traitDefinition.getName())) {
				traitDefinitionProperties = traitDefinition;
			}
		}

		while (xmlReader.hasNext()) {
			xmlReader.next();

			if (xmlReader.isStartElement()
					&& ELEMENT_ARCHITECTURE.equalsIgnoreCase(xmlReader
							.getLocalName())) {
				String frameName = xmlReader.getAttributeValue(null,
						ATTRIBUTE_FRAME);
				String implementation = xmlReader.getAttributeValue(null,
						ATTRIBUTE_IMPL);
				String name = xmlReader.getAttributeValue(null, ATTRIBUTE_NAME);
				String version = null;

				while (!xmlReader.isEndElement()
						|| !ELEMENT_ARCHITECTURE.equalsIgnoreCase(xmlReader
								.getLocalName())) {
					xmlReader.next();

					if (xmlReader.isStartElement()
							&& ELEMENT_PROPERTY
									.equals(xmlReader.getLocalName())) {
						String propertyName = xmlReader.getAttributeValue(null,
								ATTRIBUTE_NAME);
						String propertyValue = xmlReader.getAttributeValue(
								null, ATTRIBUTE_TYPE);
						properties.put(propertyName, propertyValue);
					}
				}

				comp.setName(name);

				for (TagDefinition tagDefinition : ct.getTagSetList()) {
					if (frameName != null && frameName.length() > 0
							&& TAG_FRAME.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(frameName));

						comp.getTagSetList().add(tag);
					} else if (implementation != null
							&& implementation.length() > 0
							&& TAG_IMPLEMENTATION.equals(tagDefinition
									.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(implementation));

						comp.getTagSetList().add(tag);
					} else if (version != null && version.length() > 0
							&& TAG_VERSION.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(version);

						comp.getTagSetList().add(tag);
					}
				}

				// convert frame name to URI
				URI frameURI = URI.create(frameName);

				List<Trait> traits = loadFrame(frameURI, properties, ct);
				if (!traits.isEmpty()) {
					comp.getTraitSetList().addAll(traits);
				}
			}
		}

		// add properties trait
		if (!properties.isEmpty()) {
			Trait traitProperties = factory.createTrait();
			traitProperties.setDef(traitDefinitionProperties);

			for (Object key : properties.keySet()) {
				Element e = factory.createElement();
				e.setName(key.toString());
				traitProperties.getElementSetList().add(e);
			}

			comp.getTraitSetList().add(traitProperties);
		}

		is.close();

		return comp;
	}

	/**
	 * Loads a frame XML file and process data mining
	 * 
	 * @param frameName
	 *            Frame name with sofatype protocol
	 * @param ct
	 *            Architecture {@link ComponentType}
	 * @param factory
	 *            ENTMM factory object
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private List<Trait> loadFrame(URI frameURI, Properties properties,
			ComponentType ct) throws IOException, XMLStreamException {
		List<Trait> traits = new ArrayList<Trait>();

		InputStream is = new FileInputStream(adls.get(frameURI));

		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

		TraitDefinition traitDefinitionProvided = null;
		TraitDefinition traitDefinitionRequired = null;

		// store trait definition
		for (TraitDefinition traitDefinition : ct.getTraitSetList()) {
			if (TRAIT_PROVIDED_INTERFACE.equals(traitDefinition.getName())) {
				traitDefinitionProvided = traitDefinition;
			} else if (TRAIT_REQUIRED_INTERFACE.equals(traitDefinition
					.getName())) {
				traitDefinitionRequired = traitDefinition;
			}
		}

		Trait traitProvides = factory.createTrait();
		traitProvides.setDef(traitDefinitionProvided);

		Trait traitRequires = factory.createTrait();
		traitRequires.setDef(traitDefinitionRequired);

		while (xmlReader.hasNext()) {
			xmlReader.next();

			if (xmlReader.isStartElement()
					&& (ELEMENT_PROVIDES.equals(xmlReader.getLocalName()) || ELEMENT_PROVIDED_INTERFACE
							.equalsIgnoreCase(xmlReader.getLocalName()))) {
				String iface = xmlReader.getAttributeValue(null,
						ATTRIBUTE_ITF_TYPE);
				String communicationStyle = xmlReader.getAttributeValue(null,
						ATTRIBUTE_COMM_STYLE);

				Element element = factory.createElement();
				element.setName(getClassName(getClassPath(iface)));

				// add tags: communication style, interface type
				for (TagDefinition tagDefinition : traitDefinitionProvided
						.getTagSetList()) {
					if (TAG_INTERFACE_TYPE.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(iface));

						element.getTagsetList().add(tag);
					} else if (TAG_COMMUNICATION_STYLE.equals(tagDefinition
							.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(communicationStyle);

						element.getTagsetList().add(tag);
					}
				}

				traitProvides.getElementSetList().add(element);
			} else if (xmlReader.isStartElement()
					&& (ELEMENT_REQUIRES.equals(xmlReader.getLocalName()) || ELEMENT_REQUIRED_INTERFACE
							.equalsIgnoreCase(xmlReader.getLocalName()))) {
				String iface = xmlReader.getAttributeValue(null,
						ATTRIBUTE_ITF_TYPE);
				String communicationStyle = xmlReader.getAttributeValue(null,
						ATTRIBUTE_COMM_STYLE);

				Element element = factory.createElement();
				element.setName(getClassName(getClassPath(iface)));

				for (TagDefinition tagDefinition : traitDefinitionRequired
						.getTagSetList()) {
					if (TAG_INTERFACE_TYPE.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(iface));

						element.getTagsetList().add(tag);
					} else if (TAG_COMMUNICATION_STYLE.equals(tagDefinition
							.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(communicationStyle);

						element.getTagsetList().add(tag);
					}
				}

				traitRequires.getElementSetList().add(element);
			} else if (xmlReader.isStartElement()
					&& ELEMENT_PROPERTY.equals(xmlReader.getLocalName())) {
				String propertyName = xmlReader.getAttributeValue(null,
						ATTRIBUTE_NAME);
				String propertyValue = xmlReader.getAttributeValue(null,
						ATTRIBUTE_TYPE);
				properties.put(propertyName, propertyValue);
			}
		}

		if (traitProvides.getElementSetLength() > 0) {
			traits.add(traitProvides);
		}

		if (traitRequires.getElementSetLength() > 0) {
			traits.add(traitRequires);
		}

		is.close();

		return traits;
	}

	/**
	 * Get a class path from SOFATYPE URI
	 * 
	 * @param source
	 *            SOFATYPE URI
	 * @return class path
	 */
	private String getClassPath(String value) {
		if (value == null || value.length() == 0)
			return null;

		URI uri = URI.create(value);
		if (SOFATYPE_SCHEME.equals(uri.getScheme())) {
			return uri.getAuthority();
		}

		return uri.getPath();
	}

	/**
	 * Get a name of class from a class path
	 * 
	 * @param classPath
	 *            class path
	 * @return name of class
	 */
	private String getClassName(String classPath) {
		int lastDot = classPath.lastIndexOf('.');
		if (lastDot == -1)
			return classPath;

		return classPath.substring(lastDot + 1, classPath.length());
	}

	public ArrayList<URI> getAssemblies() {
		return assemblies;
	}
}
