package cz.zcu.kiv.comav.loaders.sofa2;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.apache.log4j.Logger;

import cz.zcu.kiv.comav.core.export.loaders.ComponentLoader;
import cz.zcu.kiv.comav.entmm.Component;
import cz.zcu.kiv.comav.entmm.ComponentType;
import cz.zcu.kiv.comav.entmm.ENTMMFactory;
import cz.zcu.kiv.comav.entmm.Element;
import cz.zcu.kiv.comav.entmm.ElementElementBinding;
import cz.zcu.kiv.comav.entmm.Tag;
import cz.zcu.kiv.comav.entmm.TagDefinition;
import cz.zcu.kiv.comav.entmm.Trait;
import cz.zcu.kiv.comav.entmm.TraitDefinition;
import cz.zcu.kiv.comav.loaders.sofa2.util.Sofa2RepositoryInputStream;

/**
 * This class do all work around loading SOFA2 sofatypes in entmm data structure
 * 
 * It has to contain all the data mapping directly in the code
 * 
 * @author Martin Sloup
 * 
 */
public class Sofa2RepositoryComponentLoader extends ComponentLoader {
	private final static Logger logger = Logger
			.getLogger(Sofa2RepositoryComponentLoader.class);

	private static final String SOFAREP_SCHEME = "sofarep";

	// Common elements
	private static final String ELEMENT_PROPERTY = "property";
	private static final String ATTRIBUTE_HREF = "href";
	private static final String ATTRIBUTE_NAME = "name";
	private static final String ATTRIBUTE_TYPE = "type";

	// Assembly
	private static final String ELEMENT_SUBCOMPONENT = "subcomponent";

	// Architecture
	private static final String ELEMENT_ARCHITECTURE = "architecture";
	private static final String ELEMENT_VERSION = "version";
	private static final String ELEMENT_IMPLEMENTS = "implements";
	private static final String ATTRIBUTE_IMPLEMENTATION = "implementation";
	private static final String ATTRIBUTE_VERSION = "version";

	// Frame
	private static final String ELEMENT_PROVIDES = "provides";
	private static final String ELEMENT_PROVIDED_INTERFACE = "providedInterface";
	private static final String ELEMENT_REQUIRES = "requires";
	private static final String ELEMENT_REQUIRED_INTERFACE = "requiredInterface";
	private static final String ELEMENT_INTERFACE_TYPE = "interfaceType";
	private static final String ATTRIBUTE_COMMUNICATION_STYLE = "communicationStyle";

	// Traits
	private static final String TRAIT_PROVIDED_INTERFACE = "Provided interface";
	private static final String TRAIT_REQUIRED_INTERFACE = "Required interface";
	private static final String TRAIT_PROPERTIES = "Properties";

	// Tags
	private static final String TAG_FRAME = "frame";
	private static final String TAG_IMPLEMENTATION = "implementation";
	private static final String TAG_COMMUNICATION_STYLE = "Communication style";
	private static final String TAG_INTERFACE_TYPE = "Interface type";
	private static final String TAG_VERSION = "Version";

	/**
	 * Path to osgi model specification
	 */
	private static final String SOFA2_MODEL_PATH = "/cz/zcu/kiv/comav/loaders/sofa2/SOFA2.entmm";

	private URI repository = null;
	private ComponentType architectureComponentType = null;

	/**
	 * SOFA2 component model and attribute mapping is done here
	 */
	public Sofa2RepositoryComponentLoader(String repositoryLocation) {
		super(SOFA2_MODEL_PATH);

		repository = URI.create(repositoryLocation);
	}

	@Override
	public void load(URI[] components) throws IOException {
		//repository = getConfiguration(CONFIGURATION_REPOSITORY);

		// SOFA2 has only one component type - architecture
		architectureComponentType = model.getComponentSet()[0];

		for (URI component : components) {
			component = getFullURI(component);

			if (!SOFAREP_SCHEME.equals(component.getScheme()))
				throw new IllegalArgumentException(
						"SOFAREP scheme in URI is supported only. You have: "
								+ component);

			Component comp = loadAssembly(component);
			if (comp != null) {
				componentList.add(comp);
			}
		}
		createBindings(componentList, factory);
	}

	private void createBindings(List<Component> componentList,
			ENTMMFactory factory) {
		for (Component localComponent : componentList) {
			createBinding(localComponent, componentList,
					TRAIT_REQUIRED_INTERFACE, TRAIT_PROVIDED_INTERFACE, factory);

			if (localComponent.getComponentSetLength() > 0) {
				createBindings(localComponent.getComponentSetList(), factory);
			}
		}
	}

	private void createBinding(Component localComponent,
			List<Component> aliedComponens, String localTraitName,
			String aliedTraitName, ENTMMFactory factory) {
		for (Trait localTrait : localComponent.getTraitSet()) {
			if (localTrait.getDef().getName().equals(localTraitName)) {
				for (Element localElement : localTrait.getElementSet()) {
					for (Component aliedComponent : aliedComponens) {
						for (Trait aliedTrait : aliedComponent.getTraitSet()) {
							if (aliedTrait.getDef().getName()
									.equals(aliedTraitName)) {
								for (Element aliedElement : aliedTrait
										.getElementSet()) {
									if (localElement.getName().equals(
											aliedElement.getName())) {
										// TRAIT_REQUIRED_INTERFACE and
										// TRAIT_PROVIDED_INTERFACE must have
										// same value for TAG_INTERFACE_TYPE
										if (TRAIT_REQUIRED_INTERFACE
												.equals(localTraitName)
												&& !hasSameInterfaceType(
														localElement,
														aliedElement))
											continue;
										createElementElementBinding(
												localComponent, aliedComponent,
												localTraitName, localElement,
												aliedElement, factory);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	private boolean hasSameInterfaceType(Element localElement,
			Element aliedElement) {
		for (Tag localTag : localElement.getTagsetList()) {
			for (Tag aliedTag : aliedElement.getTagsetList()) {
				if (TAG_INTERFACE_TYPE.equals(localTag.getDef().getName())
						&& TAG_INTERFACE_TYPE.equals(aliedTag.getDef()
								.getName()) && localTag.getValue() != null
						&& localTag.getValue().equals(aliedTag.getValue())) {
					return true;
				}
			}
		}
		return false;
	}

	private void createElementElementBinding(Component localComponent,
			Component aliedComponent, String localTraitName,
			Element localElement, Element aliedElement, ENTMMFactory factory) {
		ElementElementBinding localBinding = factory
				.createElementElementBinding();
		localBinding.setLocal(localElement);
		localBinding.setAlien(aliedElement);
		localBinding.setProvided(false);
		localComponent.getBindingsList().add(localBinding);

		ElementElementBinding aliedBinding = factory
				.createElementElementBinding();
		aliedBinding.setLocal(aliedElement);
		aliedBinding.setAlien(localElement);
		aliedBinding.setProvided(true);
		aliedComponent.getBindingsList().add(aliedBinding);
	}

	/**
	 * Loads an assembly XML file and process data mining
	 * 
	 * @param assemblyURI
	 *            Assembly URI with SOFAREP scheme
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 */
	private Component loadAssembly(URI assemblyURI) throws IOException {
		Component comp = null;

		try {
			InputStream is = new Sofa2RepositoryInputStream(repository,
					assemblyURI);

			XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
			XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

			while (xmlReader.hasNext()) {
				xmlReader.next();

				if (xmlReader.isStartElement()
						&& ELEMENT_ARCHITECTURE
								.equals(xmlReader.getLocalName())) {
					String architectureName = xmlReader.getAttributeValue(null,
							ATTRIBUTE_HREF);

					// map architecture URI to filesystem
					URI architectureURI = assemblyURI.resolve(architectureName);

					comp = loadArchitecture(architectureURI);
					xmlReader.next();
				}

				// parse subcomponent
				if (xmlReader.isStartElement()
						&& ELEMENT_SUBCOMPONENT
								.equals(xmlReader.getLocalName())) {
					comp.getComponentSetList().add(
							loadSubComponent(xmlReader, assemblyURI));
				}
			}

			is.close();
		} catch (XMLStreamException e) {
			throw new IOException("Error while parsing XML file", e);
		}
		return comp;
	}

	/**
	 * Process a subcomponents from Assembly XML
	 * 
	 * @param xmlReader
	 *            {@link XMLStreamReader} object
	 * @param assemblyURI
	 *            Assembly URI with SOFAREP scheme
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private Component loadSubComponent(XMLStreamReader xmlReader,
			URI assemblyURI) throws IOException, XMLStreamException {
		Component comp = null;

		if (!xmlReader.isEndElement()) {
			while (!xmlReader.isEndElement()
					|| !ELEMENT_SUBCOMPONENT.equals(xmlReader.getLocalName())) {
				xmlReader.next();

				if (xmlReader.isStartElement()
						&& ELEMENT_ARCHITECTURE
								.equals(xmlReader.getLocalName())) {
					String architectureName = xmlReader.getAttributeValue(null,
							ATTRIBUTE_HREF);

					// convert architecture name to URI
					URI architectureURI = assemblyURI.resolve(architectureName);

					comp = loadArchitecture(architectureURI);
				}

				// parse subcomponent
				if (comp != null
						&& xmlReader.isStartElement()
						&& ELEMENT_SUBCOMPONENT
								.equals(xmlReader.getLocalName())) {
					comp.getComponentSetList().add(
							loadSubComponent(xmlReader, assemblyURI));
				}
			}
		}

		return comp;
	}

	/**
	 * Loads an architecture XML file and process data mining
	 * 
	 * @param architectureURI
	 *            Architecture URI with SOFAREP scheme
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private Component loadArchitecture(URI architectureURI) throws IOException,
			XMLStreamException {
		Component comp = factory.createComponent();
		comp.setDef(architectureComponentType);

		InputStream is = new Sofa2RepositoryInputStream(repository,
				architectureURI);

		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

		Properties properties = new Properties();
		TraitDefinition traitDefinitionProperties = null;

		// store trait definition
		for (TraitDefinition traitDefinition : architectureComponentType
				.getTraitSetList()) {
			if (TRAIT_PROPERTIES.equals(traitDefinition.getName())) {
				traitDefinitionProperties = traitDefinition;
			}
		}

		while (xmlReader.hasNext()) {
			xmlReader.next();

			if (xmlReader.isStartElement()
					&& ELEMENT_ARCHITECTURE.equalsIgnoreCase(xmlReader
							.getLocalName())) {
				String implementation = xmlReader.getAttributeValue(null,
						ATTRIBUTE_IMPLEMENTATION);
				String name = xmlReader.getAttributeValue(null, ATTRIBUTE_NAME);
				String frameName = null;
				String version = null;

				while (!xmlReader.isEndElement()
						|| !ELEMENT_ARCHITECTURE.equalsIgnoreCase(xmlReader
								.getLocalName())) {
					xmlReader.next();

					if (version == null && xmlReader.isStartElement()
							&& ELEMENT_VERSION.equals(xmlReader.getLocalName())) {
						version = xmlReader.getAttributeValue(null,
								ATTRIBUTE_VERSION);
					} else if (frameName == null
							&& xmlReader.isStartElement()
							&& ELEMENT_IMPLEMENTS.equals(xmlReader
									.getLocalName())) {
						frameName = xmlReader.getAttributeValue(null,
								ATTRIBUTE_HREF);
					} else if (xmlReader.isStartElement()
							&& ELEMENT_PROPERTY
									.equals(xmlReader.getLocalName())) {
						String propertyName = xmlReader.getAttributeValue(null,
								ATTRIBUTE_NAME);
						String propertyValue = xmlReader.getAttributeValue(
								null, ATTRIBUTE_TYPE);
						properties.put(propertyName, propertyValue);
					}
				}

				comp.setName(name);

				for (TagDefinition tagDefinition : architectureComponentType
						.getTagSetList()) {
					if (frameName != null && frameName.length() > 0
							&& TAG_FRAME.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(architectureURI
								.resolve(frameName)));

						comp.getTagSetList().add(tag);
					} else if (implementation != null
							&& implementation.length() > 0
							&& TAG_IMPLEMENTATION.equals(tagDefinition
									.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(implementation);

						comp.getTagSetList().add(tag);
					} else if (version != null && version.length() > 0
							&& TAG_VERSION.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(version);

						comp.getTagSetList().add(tag);
					}
				}

				if (frameName != null) {
					// convert frame name to URI
					URI frameURI = architectureURI.resolve(frameName);

					List<Trait> traits = loadFrame(frameURI, properties);
					if (!traits.isEmpty()) {
						comp.getTraitSetList().addAll(traits);
					}
				}
			}
		}

		// add properties trait
		if (!properties.isEmpty()) {
			Trait traitProperties = factory.createTrait();
			traitProperties.setDef(traitDefinitionProperties);

			for (Object key : properties.keySet()) {
				Element e = factory.createElement();
				e.setName(key.toString());
				traitProperties.getElementSetList().add(e);
			}

			comp.getTraitSetList().add(traitProperties);
		}

		is.close();

		return comp;
	}

	/**
	 * Loads a frame XML file and process data mining
	 * 
	 * @param frameURI
	 *            Frame URI with SOFAREP scheme
	 * @return Component
	 * @throws IOException
	 *             If I/O error occurs
	 * @throws XMLStreamException
	 *             If XML parsing error occurs
	 */
	private List<Trait> loadFrame(URI frameURI, Properties properties)
			throws IOException, XMLStreamException {
		List<Trait> traits = new ArrayList<Trait>();

		InputStream is = new Sofa2RepositoryInputStream(repository, frameURI);

		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is);

		TraitDefinition traitDefinitionProvided = null;
		TraitDefinition traitDefinitionRequired = null;

		// store trait definition
		for (TraitDefinition traitDefinition : architectureComponentType
				.getTraitSetList()) {
			if (TRAIT_PROVIDED_INTERFACE.equals(traitDefinition.getName())) {
				traitDefinitionProvided = traitDefinition;
			} else if (TRAIT_REQUIRED_INTERFACE.equals(traitDefinition
					.getName())) {
				traitDefinitionRequired = traitDefinition;
			}
		}

		Trait traitProvides = factory.createTrait();
		traitProvides.setDef(traitDefinitionProvided);

		Trait traitRequires = factory.createTrait();
		traitRequires.setDef(traitDefinitionRequired);

		while (xmlReader.hasNext()) {
			xmlReader.next();

			if (xmlReader.isStartElement()
					&& (ELEMENT_PROVIDES.equals(xmlReader.getLocalName()) || ELEMENT_PROVIDED_INTERFACE
							.equalsIgnoreCase(xmlReader.getLocalName()))) {
				String iface = null;
				String communicationStyle = xmlReader.getAttributeValue(null,
						ATTRIBUTE_COMMUNICATION_STYLE);

				while (!xmlReader.isEndElement()
						|| !ELEMENT_PROVIDED_INTERFACE.equals(xmlReader
								.getLocalName())) {
					xmlReader.next();

					if (xmlReader.isStartElement()
							&& ELEMENT_INTERFACE_TYPE
									.equalsIgnoreCase(xmlReader.getLocalName())) {
						iface = xmlReader.getAttributeValue(null,
								ATTRIBUTE_HREF);
					}
				}

				if (iface == null || iface.length() == 0)
					continue;

				// convert interface name to URI
				URI ifaceURI = frameURI.resolve(iface);

				Element element = factory.createElement();
				element.setName(getClassName(getClassPath(ifaceURI)));

				// add tags: communication style, interface type
				for (TagDefinition tagDefinition : traitDefinitionProvided
						.getTagSetList()) {
					if (TAG_INTERFACE_TYPE.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(ifaceURI));

						element.getTagsetList().add(tag);
					} else if (TAG_COMMUNICATION_STYLE.equals(tagDefinition
							.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(communicationStyle);

						element.getTagsetList().add(tag);
					}
				}

				traitProvides.getElementSetList().add(element);
			} else if (xmlReader.isStartElement()
					&& (ELEMENT_REQUIRES.equals(xmlReader.getLocalName()) || ELEMENT_REQUIRED_INTERFACE
							.equalsIgnoreCase(xmlReader.getLocalName()))) {
				String iface = null;
				String communicationStyle = xmlReader.getAttributeValue(null,
						ATTRIBUTE_COMMUNICATION_STYLE);

				while (!xmlReader.isEndElement()
						|| !ELEMENT_REQUIRED_INTERFACE.equals(xmlReader
								.getLocalName())) {
					xmlReader.next();

					if (xmlReader.isStartElement()
							&& ELEMENT_INTERFACE_TYPE
									.equalsIgnoreCase(xmlReader.getLocalName())) {
						iface = xmlReader.getAttributeValue(null,
								ATTRIBUTE_HREF);
					}
				}

				if (iface == null || iface.length() == 0)
					continue;

				// convert interface name to URI
				URI ifaceURI = frameURI.resolve(iface);

				Element element = factory.createElement();
				element.setName(getClassName(getClassPath(ifaceURI)));

				for (TagDefinition tagDefinition : traitDefinitionRequired
						.getTagSetList()) {
					if (TAG_INTERFACE_TYPE.equals(tagDefinition.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(getClassPath(ifaceURI));

						element.getTagsetList().add(tag);
					} else if (TAG_COMMUNICATION_STYLE.equals(tagDefinition
							.getName())) {
						Tag tag = factory.createTag();
						tag.setDef(tagDefinition);
						tag.setValue(communicationStyle);

						element.getTagsetList().add(tag);
					}
				}

				traitRequires.getElementSetList().add(element);
			} else if (xmlReader.isStartElement()
					&& ELEMENT_PROPERTY.equals(xmlReader.getLocalName())) {
				String propertyName = xmlReader.getAttributeValue(null,
						ATTRIBUTE_NAME);
				String propertyValue = xmlReader.getAttributeValue(null,
						ATTRIBUTE_TYPE);
				properties.put(propertyName, propertyValue);
			}
		}

		if (traitProvides.getElementSetLength() > 0) {
			traits.add(traitProvides);
		}

		if (traitRequires.getElementSetLength() > 0) {
			traits.add(traitRequires);
		}

		is.close();

		return traits;
	}

	/**
	 * Get a class path from SOFAREP URI
	 * 
	 * @param source
	 *            SOFAREP URI
	 * @return class path
	 */
	private String getClassPath(URI source) {
		return source.getPath().replace('/', '.').substring(1);
	}

	/**
	 * Get a name of class from a class path
	 * 
	 * @param classPath
	 *            class path
	 * @return name of class
	 */
	private String getClassName(String classPath) {
		int lastDot = classPath.lastIndexOf('.');
		if (lastDot == -1)
			return classPath;

		return classPath.substring(lastDot + 1, classPath.length());
	}

	/**
	 * Get a full SOFAREP URI if uri argument is not valid SOFAREP uri. E.g.
	 * convert class path to SOFAREP URI
	 * 
	 * @param uri
	 *            URI to convert
	 * @return valid SOFAREP URI
	 */
	private URI getFullURI(URI uri) {
		String authority = uri.getAuthority();
		if (authority == null)
			authority = Sofa2RepositoryInputStream.AUTHORITY_META;
		String path = uri.getPath();
		if (path != null) {
			if (!path.startsWith("/"))
				path = "/" + path;

			path = path.replace('.', '/');
		}
		try {
			return new URI(SOFAREP_SCHEME, authority, path, uri.getQuery(), "/");
		} catch (URISyntaxException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
	}

}
