package cz.zcu.kiv.comav.loaders.sofa2.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.repository.RepositoryAgent;
import org.objectweb.dsrg.sofa.repository.RepositoryFacade;
import org.objectweb.dsrg.sofa.repository.SofaURIFactory;
import org.objectweb.dsrg.sofa.repository.model.VersionedEntity;

/**
 * InputStream implementation for Sofa2 repository<br>
 * <br>
 * Based on SofaInputStream class from Sofa 2 project.
 * 
 * @author Martin Sloup
 */
public class Sofa2RepositoryInputStream extends InputStream {
	private static final Logger logger = Logger
			.getLogger(Sofa2RepositoryInputStream.class);

	/**
	 * Parameter used to determine version of content of repository
	 */
	public static final String VERSION = "version";

	/**
	 * Parameter used to determine tagged version of content of repository
	 */
	public static final String TAG = "tag";

	/**
	 * String prefix of returning error message
	 */
	public static final String ERROR_MESSAGE_PREFIX = "SOFA_ERROR_";

	/** Internal authority */
	public static final String AUTHORITY_INTERNAL = "internal";
	/** Meta authority. Store XML definition. */
	public static final String AUTHORITY_META = "meta";
	/** Data authority. Store JAR archive of implementation. */
	public static final String AUTHORITY_DATA = "data";
	
	  /**
	   * RepositoryFacade for repository lookup to resolve URIs based on tags.
	   */
	  private static RepositoryFacade tagResolver;
	  static {
	    try {
	      tagResolver = RepositoryAgent.init().getFacade();
	    } catch (Exception e) {
	      throw new RuntimeException("Unable to get repository facade for Sofa URI conversion helper.",e);
	    }
	  }

	/**
	 * Conversion Map for obtaining file extension on remote HTTP server from
	 * URI authority part.
	 */
	private static Map<String, String> authorityToExtensionMap = new HashMap<String, String>();

	/**
	 * Initializator of authorityToExtenstionMap. Adds Sofa authorities and its
	 * extensions.
	 */
	static {
		authorityToExtensionMap.put(AUTHORITY_INTERNAL, "xml");
		authorityToExtensionMap.put(AUTHORITY_META, "xml");
		authorityToExtensionMap.put(AUTHORITY_DATA, "jar");
	}

	protected InputStream is = null;

	/**
	 * Create a new instance of {@link Sofa2RepositoryInputStream} using
	 * repository URI and filename URI
	 * 
	 * @param repository
	 *            URI to repository
	 * @param fileName
	 *            filename URI with sofarep scheme
	 * @throws IOException
	 *             If I/O error occurs
	 */
	public Sofa2RepositoryInputStream(URI repository, URI fileName)
			throws IOException {
		try {
			URI uri = convertSofaURIToRequestURI(repository, fileName);
			logger.trace("Getting URL: " + uri);

			URLConnection con = uri.toURL().openConnection();
			if (con instanceof HttpURLConnection) {
				handleError((HttpURLConnection) con);
			}
			is = con.getInputStream();
		} catch (URISyntaxException e) {
			logger.error("URI syntax error: ", e);
			throw new IOException(e);
		} catch (FileNotFoundException e) {
			logger.error("File not found: ", e);
			throw e;
		}
	}

	protected void handleError(HttpURLConnection con) throws IOException {
		int result = con.getResponseCode();
		if (result != HttpURLConnection.HTTP_OK) {
			if (result == HttpURLConnection.HTTP_NOT_FOUND) {
				throw new FileNotFoundException(
						getSofaErrorTextFromBody(con.getErrorStream()));
			} else
				throw new IOException("HTTP GET method failed. Error: "
						+ result);
		}
	}

	/**
	 * This method is used to parse error response body and returns Sofa HTTP
	 * error code, because error text from server is specified only in body in
	 * response HTML text.
	 * 
	 * @param getMethod
	 *            failed getMethod from HttpClient
	 * @return Sofa error code and message and a rest of the line in stream
	 */
	private String getSofaErrorTextFromBody(InputStream is) throws IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(is));

		String line = null;
		while ((line = br.readLine()) != null) {
			if (line.contains(ERROR_MESSAGE_PREFIX)) {
				br.close();
				int i = line.indexOf(ERROR_MESSAGE_PREFIX);
				return line.substring(i);
			}
		}
		return "";
	}

	@Override
	public int read() throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		return is.read();
	}

	@Override
	public int available() throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		return is.available();
	}

	@Override
	public int read(byte[] b) throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		return is.read(b);
	}

	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		return is.read(b, off, len);
	}

	@Override
	public long skip(long n) throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		return is.skip(n);
	}

	@Override
	public synchronized void mark(int readlimit) {
		if (is != null) {
			is.mark(readlimit);
		}
	}

	@Override
	public boolean markSupported() {
		if (is == null)
			return false;
		return is.markSupported();
	}

	@Override
	public synchronized void reset() throws IOException {
		if (is == null)
			throw new IOException("Stream is not opened.");
		is.reset();
	}

	@Override
	public void close() throws IOException {
		if (is != null) {
			super.close();
			is.close();
			is = null;
		}
	}

	  /** Gets entity name from SofaURI.
	   *
	   * @param uri SofaURI
	   * @return entity name
	   */
	  private static String getEntityNameFromSofaURI(URI uri) {
		  String path = uri.getPath();
		  String entity = path.substring(1).replace('/', '.'); //path.substring(path.lastIndexOf("/")+1);
	    return entity;
	  }

	/**
	 * Converts Sofa2 repository base URI and SOFAREP URI to request URI.<br>
	 * <br>
	 * Based on a convertSofaURIToRequestURI method of a ConversionHelper class
	 * from Sofa 2 project
	 * 
	 * @param repository
	 *            repository base URI
	 * @param uri
	 *            SOFAREP URI
	 * @return request URI
	 * @throws URISyntaxException
	 *             If URI has invalid syntax
	 */
	protected URI convertSofaURIToRequestURI(URI repository, URI uri)
			throws URISyntaxException {

		StringBuffer sb = new StringBuffer();
		sb.append(repository.toString());
		// put authority
		sb.append('/');
		sb.append(uri.getAuthority());
		// hierarchical segments
		if (uri.getPath() != null)
			sb.append(uri.getPath());

		// parse query
		Map<String, String> params = parseQueryString(uri.getQuery());

		if ("data".equalsIgnoreCase(uri.getAuthority())
				&& !params.containsKey("data.stream")) {
			params.put("data.stream", "original");
		}

		String version = null;

		// version, tag and data.stream
		if (uri.getQuery() != null) {
			// Convert Tag param (if any)
			if (params.containsKey(TAG)) {
		        VersionedEntity ve = tagResolver.lookupEntityByTag(getEntityNameFromSofaURI(uri),params.get(TAG));
		              if (ve != null) {
		                if (params.containsKey(VERSION) &&
		                    !params.get("version").equals(ve.getVersion().getVersion())) {
		                  throw new IllegalArgumentException("Tag and version queries in URI " +
		                                                     uri + " refer to different objects!");
		                } else {
		                  // sufficient tag
		                  params.remove(VERSION); // remove version query (if any)
		                  version = ve.getVersion().getVersion();
		                }
		              } else {
		                throw new IllegalArgumentException("Tag in URI " + uri +
		                                                   " does not refer to any object!");
		              }
		              params.remove(TAG);
			}

			// Convert Version param (if any)
			if (params.containsKey(VERSION)) {
				version = params.get(VERSION);
				params.remove(VERSION);
			}
			// Append version string
			if (version != null) {
				sb.append("/v").append(version);
			} else {
				throw new IllegalArgumentException("URI " + uri
						+ " does not contain valid version or tag query.");
			}

			// Convert DataStream param
			String dataStream = params.get("data.stream");
			if (dataStream != null) {
				sb.append(".").append(dataStream);
				params.remove("data.stream");
			}
		}
		if (version != null) {
			// append extension
			String ext = authorityToExtensionMap.get(uri.getAuthority());
			if (ext != null && ext.length() != 0) {
				sb.append(".");
				sb.append(ext);
			}
		}

		// append query
		if (!params.isEmpty()) {
			sb.append("?");
			StringBuffer queryTemp = new StringBuffer();
			Iterator<String> paramIterator = params.keySet().iterator();
			while (paramIterator.hasNext()) {
				String paramName = paramIterator.next();
				String paramValue = params.get(paramName);
				queryTemp.append("&");
				queryTemp.append(paramName);
				queryTemp.append('=');
				queryTemp.append(paramValue);
			}
			queryTemp.deleteCharAt(0);
			sb.append(queryTemp);// remove first &
		}

		String requestURI = sb.toString();
		return new URI(requestURI);
	}

	/**
	 * Parses query string from request firstly by '&' for separating params and
	 * then by '=' for separating param name and param value. Value cannot
	 * contain another '=' character.
	 * 
	 * @param query
	 *            Query string from request
	 * @return Pairs param name and param value
	 * @throws IllegalArgumentException
	 *             If error occurs while parsing query
	 */
	public static Map<String, String> parseQueryString(String query)
			throws IllegalArgumentException {
		Map<String, String> params = new HashMap<String, String>();

		if (query == null)
			return params;

		String[] pairs = query.split("&");
		for (int i = 0; i < pairs.length; i++) {
			String[] param = pairs[i].split("=");
			if (param.length > 2)
				throw new IllegalArgumentException(
						"Invalid query parameter. Cannot contain more then one '='.");

			if (params.containsKey(param[0]))
				throw new IllegalArgumentException(
						"Duplicate parameter name in query.");

			if (param.length == 1) {
				params.put(param[0], null);
			} else {
				params.put(param[0], param[1]);
			}
		}

		return params;
	}

}
